// Python Tools for Visual Studio
// Copyright(c) Microsoft Corporation
// All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the License); you may not use
// this file except in compliance with the License. You may obtain a copy of the
// License at http://www.apache.org/licenses/LICENSE-2.0
//
// THIS CODE IS PROVIDED ON AN  *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS
// OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY
// IMPLIED WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
// MERCHANTABLITY OR NON-INFRINGEMENT.
//
// See the Apache Version 2.0 License for specific language governing
// permissions and limitations under the License.

using System;
using System.Collections.Generic;

namespace Microsoft.PythonTools.Debugger {
    // Error messages - must be kept in sync with PyDebugAttach.cpp
    public enum ConnErrorMessages {
        None,
        InterpreterNotInitialized,
        UnknownVersion,
        LoadDebuggerFailed,
        LoadDebuggerBadDebugger,
        PythonNotFound,
        TimeOut,
        CannotOpenProcess,
        OutOfMemory,
        CannotInjectThread,
        SysNotFound,
        SysSetTraceNotFound,
        SysGetTraceNotFound,
        PyDebugAttachNotFound,
        RemoteNetworkError,
        RemoteSslError,
        RemoteUnsupportedServer,
        RemoteSecretMismatch,
        RemoteAttachRejected,
        RemoteInvalidUri,
        RemoteUnsupportedTransport,
    };

    static class ConnErrorExtensions {
        private static readonly Dictionary<ConnErrorMessages, string> errMessages = new Dictionary<ConnErrorMessages, string>() {
            { ConnErrorMessages.CannotInjectThread, "Cannot create thread in debuggee process" },
            { ConnErrorMessages.CannotOpenProcess, "Cannot open process for debugging" },
            { ConnErrorMessages.InterpreterNotInitialized, "Python interpreter has not been initialized in this process" },
            { ConnErrorMessages.LoadDebuggerBadDebugger, "Failed to load debugging script (incorrect version of script?)" },
            { ConnErrorMessages.LoadDebuggerFailed, "Failed to compile debugging script" },
            { ConnErrorMessages.OutOfMemory, "Out of memory" },
            { ConnErrorMessages.PythonNotFound, "Python interpreter not found" },
            { ConnErrorMessages.TimeOut, "Timeout while attaching" },
            { ConnErrorMessages.UnknownVersion, "Unknown Python version loaded in process" },
            { ConnErrorMessages.SysNotFound, "sys module not found" },
            { ConnErrorMessages.SysSetTraceNotFound, "settrace not found in sys module" },
            { ConnErrorMessages.SysGetTraceNotFound, "gettrace not found in sys module" },
            { ConnErrorMessages.PyDebugAttachNotFound, "Cannot find PyDebugAttach.dll" },
            { ConnErrorMessages.RemoteNetworkError, "Network error while connecting to remote debugging server" },
            { ConnErrorMessages.RemoteSslError, "Could not establish a secure SSL connection to remote debugging server" },
            { ConnErrorMessages.RemoteUnsupportedServer, "Remote server is not a supported Python Tools for Visual Studio debugging server" },
            { ConnErrorMessages.RemoteSecretMismatch, "Secret specified in the Qualifier string did not match the remote secret" },
            { ConnErrorMessages.RemoteAttachRejected, "Remote debugging server rejected request to attach" },
            { ConnErrorMessages.RemoteInvalidUri, "Invalid remote debugging endpoint URI" },
            { ConnErrorMessages.RemoteUnsupportedTransport, "This remote debugging transport is not supported by this version of Windows." },
        };

        internal static string GetErrorMessage(this ConnErrorMessages attachRes) {
            string msg;
            if (!errMessages.TryGetValue(attachRes, out msg)) {
                msg = "Unknown error";
            }
            return msg;
        }
    }

    [Serializable]
    public sealed class ConnectionException : Exception {
        public ConnectionException(ConnErrorMessages error) {
            Error = error;
        }

        public ConnectionException(ConnErrorMessages error, Exception innerException)
            : base(null, innerException) {
            Error = error;
        }

        public ConnErrorMessages Error {
            get {
                return (ConnErrorMessages)Data[typeof(ConnErrorMessages)];
            }
            private set {
                Data[typeof(ConnErrorMessages)] = value;
            }
        }

        public override string Message {
            get { return Error.GetErrorMessage(); }
        }
    }
}
